import torch
import numpy as np
import matplotlib.pyplot as plt
from linear_switching.linear_switching import LinearSwitching
plt.style.use('ggplot')

def train_summary(summary: dict, save: str=None) -> None:

    fig, ax = plt.subplots(1, 3, figsize=(15, 5))

    colors = {'train' : 'C0', 'val' : 'C1'}
    linestyles = {'total' : 'solid', 'center' : 'dashed', 'outside' : 'dashdot'}

    for split in summary.keys():
        if split == 'model':
            continue
        for idx, metric in enumerate(summary[split].keys()):
            metric_words = metric.split('_')
            if 'loss' in metric_words:
                ax[0].plot(summary[split][metric], c=colors[split], ls=linestyles[metric_words[0]], label=metric_words[0])
            elif 'score' in metric_words:
                ax[1].plot(summary[split][metric], c=colors[split], ls=linestyles[metric_words[0]], label=metric_words[0])
            else:
                ax[2].plot(summary[split][metric], c=colors[split], ls=linestyles[metric_words[0]], label=metric_words[0])
    
    ax[0].set_title('loss')
    ax[0].set_ylabel('avg. NLL')
    ax[0].set_xlabel('epochs')
    ax[1].set_title('score')
    ax[1].set_ylabel(r'avg. $P(y\mid x, \theta)$')
    ax[1].set_xlabel('epochs')
    ax[2].set_title('accuracy')
    ax[2].set_xlabel('epochs')
    ax[2].set_ylabel('fraction correct')

    fig.legend(['train total', 'train center', 'train outside', 'val total', 'val center', 'val outside'], bbox_to_anchor=(1.1, 1))
    fig.suptitle('performance across training')
    fig.tight_layout()
    
    if save is not None:
        plt.savefig(save)


def plot_update_rules(model: LinearSwitching, z_min: float, z_max: float, save: str=None) -> None:

    labels = [
        's=-80, a=+1, r=1',
        's=0, a=-1, r=1',
        's=0, a=1, r=1',
        's=+80, a=-1, r=1',
        's=-80, a=-1, r=0',
        's=0, a=1, r=0',
        's=0, a=-1, r=0',
        's=+80, a=1, r=0'
    ]

    coefficients = model.coefs
    offsets = model.offsets
    x_range = np.linspace(z_min, z_max, 1000)
    fig, ax = plt.subplots(2, 4, figsize=(15, 10))



    for i in range(4):

        # First row
        ax[0, i].plot(x_range, coefficients[i] * x_range + offsets[i])
        ax[0, i].set_xlabel('$z_t$')
        ax[0, i].set_ylabel('$z_{t+1}$')
        ax[0, i].set_title(labels[i])
        ax[0, i].axvline(0, ls='--', c='black', lw=0.5)
        ax[0, i].axhline(0, ls='--', c='black', lw=0.5)
        ax[0, i].set_box_aspect(1)
        ax[0, i].set_xlim(z_min, z_max)
        ax[0, i].set_ylim(z_min, z_max)
        ax[0, i].plot(np.linspace(z_min, z_max), np.linspace(z_min, z_max), c='black', ls='--')

        # Second row
        ax[1, i].plot(x_range, coefficients[i + 4] * x_range + offsets[i + 4])
        ax[1, i].set_xlabel('$z_t$')
        ax[1, i].set_ylabel('$z_{t+1}$')
        ax[1, i].set_title(labels[i + 4])
        ax[1, i].axvline(0, ls='--', c='black', lw=0.5)
        ax[1, i].axhline(0, ls='--', c='black', lw=0.5)
        ax[1, i].set_box_aspect(1)
        ax[1, i].set_xlim(z_min, z_max)
        ax[1, i].set_ylim(z_min, z_max)
        ax[1, i].plot(np.linspace(z_min, z_max), np.linspace(z_min, z_max), c='black', ls='--')
    fig.tight_layout()
    if save is not None:
        plt.savefig(save)




import torch
import torch.distributions as dist
from torch import nn
from torch import Tensor

class MLP(nn.Module):
    def __init__(self,
                 input_size: int,
                 output_size: int,
                 hidden_size: list[int],
                 act: nn.Module=nn.ReLU) -> None:
        super(MLP, self).__init__()
        layers = []
        _prev_dim = input_size
        for hidden in hidden_size:
            layers.append(nn.Linear(_prev_dim, hidden))
            layers.append(act())
            _prev_dim = hidden
        layers.append(nn.Linear(_prev_dim, output_size))
        self.mlp = nn.Sequential(*layers)
        
            
    def forward(self, x: Tensor) -> Tensor:
        x_hat = self.mlp(x)
        return x_hat


class Bottleneck(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def compute_kld(self, mean: Tensor, var: Tensor, log_var: Tensor) -> Tensor:
        return -0.5 * torch.sum(1 + log_var - (mean ** 2) - var, dim=-1)
    
    def forward(self, x: Tensor, log_var: Tensor, multiplier: Tensor) -> Tensor:
        var = log_var.exp()
        std = var.sqrt()
        mean = torch.mul(multiplier, x)
        kld = self.compute_kld(mean, var, log_var)
        
        x_tilde = torch.randn_like(x) * std + mean

        return x_tilde, kld


class UpdateMLPs(nn.Module):
    def __init__(self, num_latents: int, in_size: int, hidden_size: list[int]) -> None:
        super().__init__()

        self.update_mlps = nn.ModuleList([MLP(in_size, 2, hidden_size) for mlp in range(num_latents)])
    
    def forward(self, x: Tensor, old_latents: Tensor):
        
        B, D, Z = x.size()

        new_latents = torch.zeros(size=(B, Z))
        for mlp_idx, mlp in enumerate(self.update_mlps):
            mlp_output = mlp(x[:, :, mlp_idx])
            u, w = mlp_output[:, 0], mlp_output[:, 1].sigmoid()
            z = old_latents[:, mlp_idx]
            new_latents[:, mlp_idx] = (1 - w) * z + u * w
        
        return new_latents


class DisRNN(nn.Module):
    def __init__(
            self,
            num_latents: int,
            num_obs: int,
            include_stimulus: bool,
            update_mlp_hidden_size: list[int],
            choice_mlp_hidden_size: list[int]
        ) -> None:
        super().__init__()

        self.update_mlp_in_size = num_latents + num_obs
        self.num_latents = num_latents

        self.z_0 = nn.Parameter(torch.randn(size=(num_latents,)))
        self.update_bottleneck_multiplier = nn.Parameter(torch.ones(size=(self.update_mlp_in_size, num_latents)))
        self.update_bottleneck_log_var = nn.Parameter(dist.Uniform(-3, -2).sample((self.update_mlp_in_size, num_latents)))
        self.global_bottleneck_multiplier = nn.Parameter(torch.ones(size=[num_latents]))
        self.global_bottleneck_log_var = nn.Parameter(dist.Uniform(-3, -2).sample([num_latents]))

        self.global_bottleneck = Bottleneck()
        self.update_bottleneck = nn.ModuleList([Bottleneck() for latent in range(num_latents)])
        self.update_mlp = UpdateMLPs(num_latents, self.update_mlp_in_size, update_mlp_hidden_size)

        if include_stimulus:
            self.choice_mlp = MLP(num_latents + 1, 2, choice_mlp_hidden_size)
        else:
            self.choice_mlp = MLP(num_latents, 2, choice_mlp_hidden_size)

    
    def forward(self, latents: Tensor, obs: Tensor, t_0: bool, s_t: Tensor=None):

        B, _ = latents.size()

        if t_0:
            latents = torch.expand_copy(self.z_0, (B, self.num_latents))
        x = torch.cat([latents, obs], dim=-1)

        update_mlp_inputs = torch.zeros(size=(B, self.update_mlp_in_size, self.num_latents))
        update_kld = torch.zeros(size=(B, self.num_latents))
        for idx, update_bottleneck in enumerate(self.update_bottleneck):
            x_tilde, kld = update_bottleneck(x, self.update_bottleneck_log_var[:, idx], self.update_bottleneck_multiplier[:, idx])
            update_mlp_inputs[:, :, idx] = x_tilde
            update_kld[:, idx] = kld

        new_latents = self.update_mlp(update_mlp_inputs, latents)
        z_tilde, global_kld = self.global_bottleneck(new_latents, self.global_bottleneck_log_var, self.global_bottleneck_multiplier)
        if s_t is not None:
            choice_mlp_input = torch.cat((z_tilde, s_t), dim=-1)
            y = self.choice_mlp(choice_mlp_input)
        else:
            y = self.choice_mlp(z_tilde)

        return y, z_tilde, global_kld, update_kld
    

    def forward_w_dropout(self, latents: Tensor, obs: Tensor, t_0: bool, global_threshold: float, update_threshold: float, s_t: Tensor=None):

        open_update_latents = (self.update_bottleneck_log_var.exp() <= update_threshold).int()
        
        open_global_latents = (self.global_bottleneck_log_var.exp() <= global_threshold).int()


        B, _ = latents.size()

        if t_0:
            latents = torch.expand_copy(self.z_0, (B, self.num_latents))
        x = torch.cat([latents, obs], dim=-1)

        update_mlp_inputs = torch.expand_copy(x.unsqueeze(-1), (B, self.update_mlp_in_size, self.num_latents))
        update_mlp_inputs *= open_update_latents

        new_latents = self.update_mlp(update_mlp_inputs, latents)
        new_latents *= open_global_latents

        if s_t is not None:
            choice_mlp_input = torch.cat((new_latents, s_t), dim=-1)
            y = self.choice_mlp(choice_mlp_input)
        else:
            y = self.choice_mlp(new_latents)

        return y, new_latents

    

    @property
    def update_sigmas(self):
        return torch.exp(self.update_bottleneck_log_var.clone().detach())
    @property
    def update_multipliers(self):
        return self.update_bottleneck_multiplier.clone().detach()

    @property
    def global_sigmas(self):
        return torch.exp(self.global_bottleneck_log_var.clone().detach())
    
    @property
    def global_multipliers(self):
        return self.global_bottleneck_multiplier.clone().detach()


import torch
from torch import nn
from torch import Tensor

class LinearSwitching(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.z_0 = nn.Parameter(torch.randn(size=(1,)))
        self._coefs = nn.Parameter(torch.ones(size=(8,)))
        self._offsets = nn.Parameter(torch.zeros(size=(8,)))
        self.inv_temp = nn.Parameter(torch.rand(size=(1,)))
    
    def forward(self, z: Tensor, obs: Tensor):
        coefs = self._coefs[obs].unsqueeze(1)
        offsets = self._offsets[obs].unsqueeze(1)
        return coefs * z + offsets
    
    @property
    def coefs(self):
        return self._coefs.detach().clone()
    
    @property
    def offsets(self):
        return self._offsets.detach().clone()


import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from ..data.data_utils import load_master_dataset, load_split


def convert_dataset(dataset: list[dict]):
    observations = {
        (-1, 1, 1) : 0,
        (0, -1, 1) : 1,
        (0, 1, 1) : 2,
        (1, -1, 1) : 3,
        (-1, -1, 0) : 4,
        (0, 1, 0) : 5,
        (0, -1, 0) : 6,
        (1, 1, 0) : 7
    }
    for session in dataset:
        X = session['X']
        # One_hot_encode
        one_hot_encoded = np.array(list(map(lambda x: observations[tuple(x)], session['X'])))
        
        one_hot_matrix = np.zeros(shape=(len(X), 8))
        one_hot_matrix[np.arange(len(X)), one_hot_encoded] = 1
        sliding = sliding_window_view(one_hot_matrix, 10, 0)
        session['X'] = sliding[:-1]
        session['y'] = session['y'][10:]
    return dataset


class LogisticRegressionDataset:
    def __init__(
            self,
            subject: str,
            split: str,
            fold: int
    ) -> None:
        
        full_dataset = load_master_dataset(subject)

        self.sessions = load_split(subject, fold)[split]
        self.dataset = list(filter(lambda x: x['session_id'] in self.sessions, full_dataset))
        self.center_stim_trials = list(map(lambda session: np.where(session['X'][10:, 0] == 0), self.dataset))
        self.outside_stim_trials = list(map(lambda session: np.where(session['X'][10:, 0] != 0), self.dataset))
        self.dataset = convert_dataset(self.dataset)
    
    
    def build_dataset(self) -> tuple[np.ndarray, np.ndarray]:

        X = [self.dataset[session]['X'][self.center_stim_trials[session]] for session in range(len(self.dataset))]
        X = np.vstack(X).reshape(-1, 80)
        y = [self.dataset[session]['y'][self.center_stim_trials[session]] for session in range(len(self.dataset))]
        y = np.hstack(y)
        return X, y


    def get_session(self, session_id: int) -> dict:
        
        for session in self.dataset:
            if session['session_id'] == session_id:
                return session
        raise KeyError('invalid session ID for this dataset')
    



from disrnn.disrnn import DisRNN
from disrnn.disrnn_utils import DisRNNDataset
from data.data_utils import custom_collate_fn
import torch
from torch import Tensor
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import DataLoader



def train(
        subject: str,
        fold: int,
        epochs: int,
        lr: float,
        beta_weight: float
) -> None:
    
    # Setup
    model = DisRNN(
        num_latents=5,
        num_obs=3,
        include_stimulus=True,
        update_mlp_hidden_size=[5, 5, 5],
        choice_mlp_hidden_size=[3, 3]
    )
    model = torch.compile(model)
    optimizer = optim.Adam(model.parameters(), lr=lr)
    train_set = DisRNNDataset(subject, 'train', fold)
    val_set = DisRNNDataset(subject, 'val', fold)
    train_loader = DataLoader(
        train_set,
        batch_size=8,
        collate_fn=custom_collate_fn
    )
    val_loader = DataLoader(
        val_set,
        batch_size=8,
        collate_fn=custom_collate_fn
    )

    summary = {
        split : {
            'loss': {
                'nll' : {
                    'total' : [],
                    'c_stim' : [],
                    'o_stim' : []
                },
                'kld' : {
                    'total' : [],
                    'c_stim' : [],
                    'o_stim' : []
                }
            },
            'score' : {
                'total' : [],
                'c_stim' : [],
                'o_stim' : []
            },
            'acc' : {
                'total' : [],
                'c_stim' : [],
                'o_stim' : []
            }

        }
        for split in ['train', 'val']
    }

    summary['model'] = {
        'global_sigmas' : [],
        'global_multipliers' : [],
        'update_sigmas' : [],
        'update_multipliers' : []
    }

    summary['grad_norm'] = []
    model_checkpoint = None


    # Begin training
    for epoch in range(epochs):
        for split in ['train', 'val']:
            if split == 'val':
                torch.set_grad_enabled(False)
                model.eval()
                loader = val_loader
                dataset = val_set
            else:
                torch.set_grad_enabled(True)
                model.train()
                loader = train_loader
                dataset = train_set
            
            epoch_c_stim_nll = 0
            epoch_o_stim_nll = 0

            epoch_c_stim_kld = 0
            epoch_o_stim_kld = 0

            epoch_c_stim_score = 0
            epoch_o_stim_score = 0
            
            epoch_c_stim_correct = 0
            epoch_o_stim_correct = 0


            for (X, y, mask) in loader:
                X = X.float()
                y = y.long()
                optimizer.zero_grad()
                if split == 'train':
                    summary['model']['global_sigmas'].append(model.global_sigmas)
                    summary['model']['global_multipliers'].append(model.global_multipliers)
                    summary['model']['update_sigmas'].append(model.update_sigmas)
                    summary['model']['update_multipliers'].append(model.update_multipliers)

                B, T, _ = X.shape

                # Initialize containers for loss
                cumulated_logits = torch.zeros(size=(B, T - 1, 2))
                cumulated_kld_loss = torch.zeros(size=(B, T - 1))

                # Iterate over timesteps
                # Forward pass
                for t in range(T - 1):
                    obs = X[:, t]
                    s_t = X[:, t + 1, 0].unsqueeze(-1)
                    if t == 0:
                        dummy_z0 = torch.zeros(size=(B, model.num_latents))
                        logits, updated_z, global_kld, update_kld = model.forward(
                            dummy_z0, obs, True, s_t
                        )
                    else:
                        logits, updated_z, global_kld, update_kld = model.forward(
                            updated_z, obs, False, s_t
                        )

                    cumulated_logits[:, t] = logits
                    global_kld = global_kld.sum()
                    update_kld = update_kld.sum()
                    cumulated_kld_loss[:, t] = global_kld + update_kld

                # Flatten and apply mask
                X = X[:, 1:, :].reshape(-1, 3)
                y = y.view(-1)
                cumulated_logits = cumulated_logits.view(-1, 2)
                kld_loss = cumulated_kld_loss.view(-1)
                mask = mask.view(-1)

                # Calculate cross entropy and apply mask to loss
                probs = cumulated_logits.softmax(-1)
                log_probs = probs.log()
                nll = F.nll_loss(log_probs, y, reduction='none')
                nll[mask == False] = 0
                kld_loss[mask == False] = 0

                # Backward pass
                loss = (nll + kld_loss * beta_weight).mean()
                print(split)
                if split == 'train':
                    loss.backward()
                    optimizer.step()

                # Get trial identity indices
                center_stim_trials = torch.where((X[:, 0] == 0) & (mask == True))
                outside_stim_trials = torch.where((X[:, 0] != 0) & (mask == True))

                # Filter loss by trial type
                epoch_c_stim_nll += nll[center_stim_trials].sum().item()
                epoch_o_stim_nll += nll[outside_stim_trials].sum().item()
                epoch_c_stim_kld += kld_loss[center_stim_trials].sum().item() * beta_weight
                epoch_o_stim_kld += kld_loss[outside_stim_trials].sum().item() * beta_weight

                # Compute score
                correct_probs = probs[torch.arange(len(probs)), y]
                epoch_c_stim_score += correct_probs[center_stim_trials].sum().item()
                epoch_o_stim_score += correct_probs[outside_stim_trials].sum().item()

                # Compute correct trials
                choice = probs.argmax(-1)
                correct_choices = (choice == y)
                epoch_c_stim_correct += correct_choices[center_stim_trials].sum()
                epoch_o_stim_correct += correct_choices[outside_stim_trials].sum()

            

            summary[split]['loss']['nll']['total'].append((epoch_c_stim_nll + epoch_o_stim_nll) / dataset.length)
            summary[split]['loss']['nll']['c_stim'].append(epoch_c_stim_nll / dataset.c_stim_trials)
            summary[split]['loss']['nll']['o_stim'].append(epoch_o_stim_nll / dataset.o_stim_trials)

            summary[split]['loss']['kld']['total'].append((epoch_c_stim_kld + epoch_o_stim_kld) / dataset.length)
            summary[split]['loss']['kld']['c_stim'].append(epoch_c_stim_kld / dataset.c_stim_trials)
            summary[split]['loss']['kld']['o_stim'].append(epoch_o_stim_kld / dataset.o_stim_trials)

            summary[split]['score']['total'].append((epoch_c_stim_score + epoch_o_stim_score) / dataset.length)
            summary[split]['score']['c_stim'].append(epoch_c_stim_score / dataset.c_stim_trials)
            summary[split]['score']['o_stim'].append(epoch_o_stim_score / dataset.o_stim_trials)

            summary[split]['acc']['total'].append((epoch_c_stim_correct + epoch_o_stim_correct) / dataset.length)
            summary[split]['acc']['c_stim'].append(epoch_c_stim_correct / dataset.c_stim_trials)
            summary[split]['acc']['o_stim'].append(epoch_o_stim_correct / dataset.o_stim_trials)

        print('=' * 30)
        print(f'Epoch {epoch + 1}')
        print('Training')
        print(f"NLL: {summary['train']['loss']['nll']['total'][-1]:0.4f}, KLD: {summary['train']['loss']['kld']['total'][-1]:0.4f}")
        print(f"Score: {summary['train']['score']['total'][-1]:0.4f}, Accuracy: {summary['train']['acc']['total'][-1]}")
        print('-' * 15)
        print('Validation')
        print(f"NLL: {summary['val']['loss']['nll']['total'][-1]:0.4f}, KLD: {summary['val']['loss']['kld']['total'][-1]:0.4f}")
        print(f"Score: {summary['val']['score']['total'][-1]:0.4f}, Accuracy: {summary['val']['acc']['total'][-1]}")
    return summary


























import numpy as np
import pickle
import os
import yaml
from torch.nn.utils.rnn import pad_sequence


def load_behavioral_data(subject: str) -> list[np.ndarray]:
    '''
    Helper function for loading behavioral data CSV file for a specific mouse
    and separating their data by sessions.
    '''

    csv_path = os.path.join(os.curdir, 'subjects', subject, f'{subject}behavioralData.csv')
    dataset = np.loadtxt(csv_path, delimiter=',')
    
    unique_sessions = np.unique(dataset[:, -1])
    split_by_session = [
        (session, dataset[dataset[:, -1] == session]) for session in unique_sessions
    ]
    return split_by_session


def filter_data(dataset: list[np.ndarray]) -> list[np.ndarray]:
    '''
    Filters down dataset. Each session must have greater than 200 trials, 60%
    or higher performance on the center stim trials for each block, and 80%
    or higher performance on the outside stim trials for each block.
    '''
    def check_outside_stim():
        l_block_outside_stim = outside_stim_trials[outside_stim_trials[:, 3] == -1]
        r_block_outside_stim = outside_stim_trials[outside_stim_trials[:, 3] == 1]

        l_block_correct = (l_block_outside_stim[:, 2] == 1).sum()
        r_block_correct = (r_block_outside_stim[:, 2] == 1).sum()

        if len(outside_stim_trials) == 0:
            return False
        return (l_block_correct / len(l_block_outside_stim)) > 0.8 and \
               (r_block_correct / len(r_block_outside_stim)) > 0.8
    
    def check_center_stim():
        l_block_center_stim = center_stim_trials[center_stim_trials[:, 3] == -1]
        r_block_center_stim = center_stim_trials[center_stim_trials[:, 3] == 1]
        if len(l_block_center_stim) == 0:
            return False
        elif len(r_block_center_stim) == 0:
            return False
        else:
            l_block_correct = (l_block_center_stim[:, 2] == 1).sum()
            r_block_correct = (r_block_center_stim[:, 2] == 1).sum()
        return (l_block_correct / len(l_block_center_stim)) > 0.6 and \
               (r_block_correct / len(r_block_center_stim)) > 0.6


    new_dataset = []

    for (session_id, data) in dataset:
        if len(data) < 200:
            continue
        outside_stim_trials = data[data[:, 0] != 0]
        center_stim_trials = data[data[:, 0] == 0]
        if check_outside_stim() and check_center_stim():
            new_dataset.append((session_id, data))
    return new_dataset



def organize_data(dataset: list) -> list:
    '''
    Takes a data list where each list is a np array of the behavioral data from a
    different session and organizes it into a dictionary that allows easy access
    of input/output pairs and other info about sessions.
    '''

    organized_dataset = []

    for (session_id, data) in dataset:
        organized_dataset.append(
            {
                'session_id' : session_id,
                'X' : data[:, :3],
                'y' : data[:, 1],
                'block_label' : data[:, 3]
            }
        )
    
    return organized_dataset


def save_master_dataset(subject: str, dataset: list) -> None:
    '''
    Builds master datasets for all subjects
    '''
    master_dataset_path = os.path.join(os.curdir, 'subjects', subject, 'master_dataset.pkl')
    with open(master_dataset_path, 'wb') as f:
        pickle.dump(dataset, f)
    print(f'Saved master dataset for subject {subject} at {master_dataset_path}')


def build_master_datasets_for_subjects(subjects: list[str]) -> None:
    for subject in subjects:
        behavioral_data = organize_data(filter_data(load_behavioral_data(subject)))
        save_master_dataset(subject, behavioral_data)

def load_master_dataset(subject: str) -> dict:
    path = os.path.join(
        os.curdir,
        'data',
        'subjects',
        subject,
        'master_dataset.pkl'
    )
    with open(path, 'rb') as file:
        dataset = pickle.load(file)
    
    return dataset



def make_tvt_split(dataset: list[dict], split_ratios: tuple=(0.7, 0.15, 0.15)):
    num_sessions = len(dataset)
    train_ratio, val_ratio, _ = split_ratios
    train_end = int(num_sessions * train_ratio)
    val_end = train_end + int(num_sessions * val_ratio)
    
    
    sessions = list(map(lambda x: int(x['session_id']), dataset))
    np.random.shuffle(sessions)
    train_sessions = sessions[:train_end]
    val_sessions = sessions[train_end:val_end]
    test_sessions = sessions[val_end:]

    split = {
        'train' : train_sessions,
        'val' : val_sessions,
        'test' : test_sessions
    }
    
    return split


def build_subject_splits(subjects: list[str]):

    path = os.path.join(
        os.curdir,
        'data',
        'subjects'
    )

    for subject in subjects:
        split_path = os.path.join(path, subject, 'splits')
        dataset = load_master_dataset(subject)
        for idx, split in enumerate(range(5)):
            split = make_tvt_split(dataset)
            print(split)
            split_file = os.path.join(split_path, f'split_{idx + 1}.yaml')
            with open(split_file, 'w') as f:
                yaml.dump(split, f)


def load_split(subject: str, fold: int) -> dict:
    path = os.path.join(
        os.curdir,
        'data',
        'subjects',
        subject,
        'splits',
        f'split_{fold}.yaml'
    )
    with open(path, 'r') as f:
        session_split = yaml.safe_load(f)
    return session_split

def custom_collate_fn(batch):
    X, y = zip(*batch)
    
    padded_X = pad_sequence(X, batch_first=True, padding_value=-999)
    padded_y = pad_sequence(y, batch_first=True, padding_value=-999)
    mask = (padded_y != -999)
    padded_X[padded_X == -999] = 0
    padded_y[padded_y == -999] = 0
    return padded_X, padded_y, mask


from data.data_utils import load_master_dataset, load_split
import torch
from torch import Tensor
from torch.utils.data import Dataset
import numpy as np
from torch.nn.utils.rnn import pad_sequence

class DisRNNDataset(Dataset):
    def __init__(self, subject: str, split: str, fold: int) -> None:
        super().__init__()
        full_dataset = load_master_dataset(subject)
        self.split = split
        self.fold = load_split(subject, fold)
        self.sessions = self.fold[split]
        self.dataset = list(filter(lambda x: x['session_id'] in self.sessions, full_dataset))

        self.length = sum(map(lambda x: len(x['y']), self.dataset))
        self.c_stim_trials = sum(map(lambda x: np.sum(x['X'][1:, 0] == 0), self.dataset))
        self.o_stim_trials = sum(map(lambda x: np.sum(x['X'][1:, 0] != 0), self.dataset))
        self.error_trials = sum(map(lambda x: np.sum(x['X'][1:, 2] == 0), self.dataset))

    def __getitem__(self, index) -> Tensor:

        session = self.dataset[index]
        X, y = session['X'], session['y'][1:]
        y[y == -1] = 0

        return torch.from_numpy(X), torch.from_numpy(y)
    
    def __len__(self) -> int:
        return len(self.dataset)
    

def count_entries(d):
    return np.sum(d['X'][:, 0] == 0)


import torch
import torch.nn.functional as F
from torch import nn
from torch import optim
from torch.utils.data import DataLoader
from linear_switching.linear_switching import LinearSwitching
from linear_switching.linear_switching_utils import LinearSwitchingDataset
from ..data.data_utils import custom_collate_fn


def train(
        subject: str,
        fold: int,
        epochs: int,
        lr: float
) -> dict:
    

    model = LinearSwitching()
    optimizer = optim.Adam(model.parameters(), lr=lr)

    train_set = LinearSwitchingDataset(subject, 'train', fold)
    val_set = LinearSwitchingDataset(subject, 'val', fold)
    
    train_loader = DataLoader(train_set, 8, collate_fn=custom_collate_fn)
    val_loader = DataLoader(val_set, 8, collate_fn=custom_collate_fn)
    stopping_criteria = 1e-4





    summary = {
        split : {
            'total_loss' : [],
            'center_stim_loss' : [],
            'outside_stim_loss' : [],
            'total_score' : [],
            'center_stim_score' : [],
            'outside_stim_score' : [],
            'total_acc' : [],
            'center_stim_acc' : [],
            'outside_stim_acc' : []
        } for split in [
            'train',
            'val'
        ]
    }

    summary['model'] = {'coefficients' : [], 'offsets' : []}


    # Data observation mapping
    left_stim = torch.tensor([0, 4])
    center_stim = torch.tensor([1, 2, 5, 6])
    right_stim = torch.tensor([3, 7])

    # Iterate over epochs
    for epoch in range(epochs):

        for split in ['train', 'val']:
            if split == 'val':
                torch.set_grad_enabled(False)
                model.eval()
                loader = val_loader
            else:
                torch.set_grad_enabled(True)
                model.train()
                loader = train_loader

            # Initialize metrics
            total_loss = 0
            outside_stim_loss = 0
            center_stim_loss = 0
            total_count = 0
            outside_stim_count = 0
            center_stim_count = 0
            total_correct = 0
            outside_stim_correct = 0
            center_stim_correct = 0
            probs = []
            outside_stim_probs = []
            center_stim_probs = []

            # Iterate over batches
            for (X, y, mask) in loader:
                if split == 'train':
                    summary['model']['coefficients'].append(model.coefs)
                    summary['model']['offsets'].append(model.offsets)
                y[y == -1] = 0

                B, T = X.shape
                optimizer.zero_grad()
                cumulated_log_probs = torch.zeros(size=(B, T - 1, 2))
                z = model.z_0.expand(size=(B, 1))

                for t in range(T - 1):
                    z = model(z, X[:, t])
                    l_stim_trials = torch.where(torch.isin(X[:, t + 1], left_stim))[0]
                    r_stim_trials = torch.where(torch.isin(X[:, t + 1], right_stim))[0]
                    c_stim_trials = torch.where(torch.isin(X[:, t + 1], center_stim))[0]

                    log_probs = (1 / model.inv_temp.exp()) * z[c_stim_trials, 0]
                    p = torch.sigmoid(log_probs)
                    cumulated_log_probs[c_stim_trials, t, 0] = p.log()
                    cumulated_log_probs[c_stim_trials, t, 1] = (1 - p).log()

                cumulated_log_probs = cumulated_log_probs.view(-1, 2)
                y = y.view(-1)
                mask = mask.view(-1)
                X = X[:, 1:].reshape(-1)[mask]

                l_stim_trials = torch.where(torch.isin(X, left_stim))[0]
                r_stim_trials = torch.where(torch.isin(X, right_stim))[0]
                c_stim_trials = torch.where(torch.isin(X, center_stim))[0]

                total_count += mask.sum().item()
                outside_stim_count += len(l_stim_trials) + len(r_stim_trials)
                center_stim_count += len(c_stim_trials)

                loss = F.nll_loss(cumulated_log_probs, y.long(), reduction='none')
                masked_loss = loss * mask
                masked_loss[l_stim_trials] = 0
                masked_loss[r_stim_trials] = 0

                total_loss += masked_loss.sum().item()
                outside_stim_loss += (masked_loss[l_stim_trials].sum() + masked_loss[r_stim_trials].sum()).item()
                center_stim_loss += masked_loss[c_stim_trials].sum().item()

                if split == 'train':
                    masked_loss.mean().backward()
                    optimizer.step()

                probabilities = cumulated_log_probs[mask].exp()
                probabilities[l_stim_trials, 1] = 1
                probabilities[l_stim_trials, 0] = 0
                probabilities[r_stim_trials, 0] = 1
                probabilities[r_stim_trials, 1] = 0

                choices = probabilities.argmax(-1) == y[mask]
                outside_stim_correct += torch.cat([choices[l_stim_trials], choices[r_stim_trials]]).sum().item()
                center_stim_correct += choices[c_stim_trials].sum().item()
                total_correct += choices.sum().item()

                probabilities = probabilities[torch.arange(len(probabilities)), y[mask].long()]
                probs.append(probabilities)
                outside_stim_probs.append(torch.cat([probabilities[l_stim_trials], probabilities[r_stim_trials]]))
                center_stim_probs.append(probabilities[c_stim_trials])

            summary[split]['total_acc'].append(total_correct / total_count)
            summary[split]['center_stim_acc'].append(center_stim_correct / center_stim_count)
            summary[split]['outside_stim_acc'].append(outside_stim_correct / outside_stim_count)
            summary[split]['total_loss'].append(total_loss / total_count)
            summary[split]['center_stim_loss'].append(center_stim_loss / center_stim_count)
            summary[split]['outside_stim_loss'].append(outside_stim_loss / outside_stim_count)

            probs = torch.cat(probs)
            outside_stim_probs = torch.cat(outside_stim_probs)
            center_stim_probs = torch.cat(center_stim_probs)

            summary[split]['total_score'].append(probs.mean().item())
            summary[split]['center_stim_score'].append(center_stim_probs.mean().item())
            summary[split]['outside_stim_score'].append(outside_stim_probs.mean().item())

        print(f'Epoch {epoch + 1}')
        print(f"Training loss: {summary['train']['total_loss'][-1]:0.4f}, Training score: {summary['train']['total_score'][-1]:0.4f}, Training acc: {summary['train']['total_acc'][-1]:0.4f}")
        print(f"Validation loss: {summary['val']['total_loss'][-1]:0.4f}, Validation score: {summary['val']['total_score'][-1]:0.4f}, Validation acc: {summary['val']['total_acc'][-1]:0.4f}")

        if epoch > 1 and abs(summary['val']['total_loss'][-1] - summary['val']['total_loss'][-2]) < stopping_criteria:
            summary['model']['model'] = model
            return summary
    summary['model']['model'] = model
    return summary



import os
from data_utils import build_master_datasets_for_subjects

def main():
    data_path = os.path.join(os.curdir, 'subjects')
    all_subjects = os.listdir(data_path)
    build_master_datasets_for_subjects(all_subjects)

if __name__ == '__main__':
    main()


from ..data.data_utils import load_master_dataset, load_split
import torch
from torch import Tensor
import numpy as np
from torch.utils.data import Dataset
from linear_switching.linear_switching import LinearSwitching
from torch.utils.data import DataLoader
import torch.nn.functional as F


def convert_dataset(dataset: list[dict]):
    observations = {
        (-1, 1, 1) : 0,
        (0, -1, 1) : 1,
        (0, 1, 1) : 2,
        (1, -1, 1) : 3,
        (-1, -1, 0) : 4,
        (0, 1, 0) : 5,
        (0, -1, 0) : 6,
        (1, 1, 0) : 7
    }
    for session in dataset:
        X = session['X']
        session['X'] = np.array(list(map(lambda x: observations[tuple(x)], session['X'])))
    return dataset


class LinearSwitchingDataset(Dataset):
    def __init__(self, subject: str, split: str, fold: int) -> None:
        super().__init__()
        full_dataset = load_master_dataset(subject)
        self.split = split
        self.fold = load_split(subject, fold)
        self.sessions = self.fold[split]
        self.dataset = list(filter(lambda x: x['session_id'] in self.sessions, full_dataset))
        self.dataset = convert_dataset(self.dataset)

    def __getitem__(self, index) -> Tensor:
        session = self.dataset[index]
        X, y = session['X'], session['y'][1:]
        y[y == -1] = 0
        return torch.from_numpy(X), torch.from_numpy(y)
        
    def __len__(self) -> int:
        return len(self.dataset)
    

def evaluate(model: LinearSwitching, loader: DataLoader) -> dict:
    summary = {
        'total_loss' : [],
        'center_stim_loss' : [],
        'outside_stim_loss' : [],
        'total_score' : [],
        'center_stim_score' : [],
        'outside_stim_score' : [],
        'total_acc' : [],
        'center_stim_acc' : [],
        'outside_stim_acc' : []
    }
    # Data observation mapping
    left_stim = torch.tensor([0, 4])
    center_stim = torch.tensor([1, 2, 5, 6])
    right_stim = torch.tensor([3, 7])
    # Initialize metrics
    total_loss = 0
    outside_stim_loss = 0
    center_stim_loss = 0
    total_count = 0
    outside_stim_count = 0
    center_stim_count = 0
    total_correct = 0
    outside_stim_correct = 0
    center_stim_correct = 0
    probs = []
    outside_stim_probs = []
    center_stim_probs = []
    # Iterate over batches
    for (X, y, mask) in loader:

        y[y == -1] = 0
        B, T = X.shape
        cumulated_log_probs = torch.zeros(size=(B, T - 1, 2))
        z = model.z_0.expand(size=(B, 1))

        for t in range(T - 1):
            z = model(z, X[:, t])
            l_stim_trials = torch.where(torch.isin(X[:, t + 1], left_stim))[0]
            r_stim_trials = torch.where(torch.isin(X[:, t + 1], right_stim))[0]
            c_stim_trials = torch.where(torch.isin(X[:, t + 1], center_stim))[0]

            # For each trial, if center-stim transform latent into log probability
            # Outside stim trials are log(1)
            log_probs = model.temperature.exp() * z[c_stim_trials, 0]
            p = F.sigmoid(log_probs)
            cumulated_log_probs[c_stim_trials, t, 0] = p.log()
            cumulated_log_probs[c_stim_trials, t, 1] = (1 - p).log()

        # Flatten log probabilities
        cumulated_log_probs = cumulated_log_probs.view(-1, 2)
        y = y.view(-1)
        mask = mask.view(-1)
        X = X[:, 1:].reshape(-1)[mask == True]

        # Indices of different trial types
        l_stim_trials = torch.where(torch.isin(X, left_stim))[0]
        r_stim_trials = torch.where(torch.isin(X, right_stim))[0]
        c_stim_trials = torch.where(torch.isin(X, center_stim))[0]

        # Counts for different trial types
        total_count += mask.sum().item()
        outside_stim_count += len(l_stim_trials) + len(r_stim_trials)
        center_stim_count += len(c_stim_trials)

        # Calculate loss on legit trials
        loss = F.nll_loss(cumulated_log_probs, y.long(), reduction='none')
        masked_loss = loss * mask
        masked_loss[l_stim_trials] = 0
        masked_loss[r_stim_trials] = 0

        # Accumulate loss for different trial types
        total_loss += masked_loss.sum().item()
        outside_stim_loss += (masked_loss[l_stim_trials].sum() + masked_loss[r_stim_trials].sum()).item()
        center_stim_loss += masked_loss[c_stim_trials].sum().item()

        # Accumulate correct trials for computing accuracy
        probabilities = cumulated_log_probs[mask == True].exp()
        probabilities[l_stim_trials, 1] = 1
        probabilities[l_stim_trials, 0] = 0
        probabilities[r_stim_trials, 0] = 1
        probabilities[r_stim_trials, 1] = 0
        choices = probabilities.argmax(-1)
        choices = choices == y[mask == True]
        outside_stim_correct += torch.cat([choices[l_stim_trials], choices[r_stim_trials]]).sum()
        center_stim_correct += choices[c_stim_trials].sum()
        total_correct += choices.sum()
        probabilities = probabilities[torch.arange(len(probabilities)), y[mask == True].int()]
        probs.append(probabilities)
        outside_stim_probs.append(torch.cat([probabilities[l_stim_trials], probabilities[r_stim_trials]]))
        center_stim_probs.append(probabilities[c_stim_trials])

    summary['total_acc'].append(total_correct / total_count)
    summary['center_stim_acc'].append(center_stim_correct / center_stim_count)
    summary['outside_stim_acc'].append(outside_stim_correct / outside_stim_count)

    summary['total_loss'].append(total_loss / total_count)
    summary['center_stim_loss'].append(center_stim_loss / center_stim_count)
    summary['outside_stim_loss'].append(outside_stim_loss / outside_stim_count)

    probs = torch.cat(probs)
    outside_stim_probs = torch.cat(outside_stim_probs)
    center_stim_probs = torch.cat(center_stim_probs)
    summary['total_score'].append(probs.mean().item())
    summary['center_stim_score'].append(center_stim_probs.mean().item())
    summary['outside_stim_score'].append(outside_stim_probs.mean().item())

    return summary

def latent(model: LinearSwitching, dataset: LinearSwitchingDataset, session: int) -> np.ndarray:
    zs = []
    X, y = dataset[session]
    with torch.no_grad():
        T = len(X)
        z = model.z_0.unsqueeze(0)
        X = X.unsqueeze(0)
        for t in range(T - 1):
            zs.append(z)
            z = model(z, X[:, t])
    zs = torch.stack(zs).detach()
    return zs[:, 0, 0]